/**
 *
 *    Copyright (C) 2017 J. Robert (FAU Erlangen-Nuernberg)
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 *    \brief This file contains the implementation for the required filter functions.
 *
 */


#include <math.h>
#include <stdlib.h>

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif


#include "ieee_lpwa_filter.h"
#include "ieee_lpwa_noise.h"


long int root_raised_cosine_filter(double **filter_pp, const double bandwidth,
        const double alpha, const long int num_taps) {

    *filter_pp = (double*)malloc(sizeof(double) * num_taps);

    if (*filter_pp == NULL)
        return -1;


    //////////////////////////////////
    // Calculate coefficients
    // of raised cosine filter
    //////////////////////////////////
    long int i;
    for (i = 0; i < num_taps; i++) {

        // t is scaled wrt. the bandwidth
        double t = (double)(-num_taps / 2 + i) * bandwidth;

        // special handling for t == 0
        if (fabs(t) < 1.0e-25) {
            (*filter_pp)[i] = 1.0 - alpha  + (4.0 / M_PI) * alpha;
            continue;
        }

        // special handling for abs(t) == 1 / (4 * alpha)
        if (fabs(fabs(t) - 1.0 / (4.0 * alpha)) < 1.0e-25) {
            (*filter_pp)[i] = alpha * 0.5 * (
                    (1.0 + M_PI / 2.0) * sin(M_PI / (4.0 * alpha))
                    + (1.0 - 2.0 / M_PI) * cos (M_PI / (4.0 * alpha))
                    );
            continue;
        }

        // all the other cases
        const double num = sin(M_PI * (1.0 - alpha) * t)
                + 4.0 * alpha * t * cos (M_PI * (1.0 + alpha) * t);
   
        const double denom = M_PI * t * (1.0 - (4.0 * alpha) * (4.0 * alpha) * t * t);

        (*filter_pp)[i] = num / denom;
    }

    //////////////////////////////////
    // Normalize filter to power one
    //////////////////////////////////
    double energy = 0;
    for (i = 0; i < num_taps; i++) {
        energy += (*filter_pp)[i] * (*filter_pp)[i];
    }

    const double correction = 1.0 / sqrt(energy);
    for (i = 0; i < num_taps; i++) {
        (*filter_pp)[i] *= correction;
    }


    return num_taps;

}

long int upsample(const double *in_re_p, const double *in_im_p,
        double **out_re_pp, double **out_im_pp,
        const long int len) {

    /*
     * The halfband filter bases on the 39 tap filter the following
     * filter coefficients
     *
     *  -1.61366968e-04,   0.00000000e+00,   8.00551641e-04,
     *   0.00000000e+00,  -2.57943188e-03,   0.00000000e+00,
     *   6.58749938e-03,   0.00000000e+00,  -1.44654955e-02,
     *   0.00000000e+00,   2.86479117e-02,   0.00000000e+00,
     *  -5.32041478e-02,   0.00000000e+00,   9.72224271e-02,
     *   0.00000000e+00,  -1.92736874e-01,   0.00000000e+00,
     *   6.29878442e-01,   1.00000000e+00,   6.29878442e-01,
     *   0.00000000e+00,  -1.92736874e-01,   0.00000000e+00,
     *   9.72224271e-02,   0.00000000e+00,  -5.32041478e-02,
     *   0.00000000e+00,   2.86479117e-02,   0.00000000e+00,
     *  -1.44654955e-02,   0.00000000e+00,   6.58749938e-03,
     *   0.00000000e+00,  -2.57943188e-03,   0.00000000e+00,
     *   8.00551641e-04,   0.00000000e+00,  -1.61366968e-04
     */


    // Filter coefficients without zeros and only positive time axis
    const double halfBandFilterCoeff[] = {
            6.29878442e-01, -1.92736874e-01,
            9.72224271e-02, -5.32041478e-02,
            2.86479117e-02, -1.44654955e-02,
            6.58749938e-03, -2.57943188e-03,
            8.00551641e-04, -1.61366968e-04};


    // Number of coefficients of the halfband filter
    const long int NUM_HB_COEFF = sizeof(halfBandFilterCoeff) / sizeof(double);

    // Length of the halfband filter if realized as classical FIR filter
    const int NUM_TAPS = NUM_HB_COEFF * 4 - 1;

    // Length of output data is twice the length of the filtered signal
    const long int output_length = 2 * (len + NUM_TAPS - 1);

    ///////////////////////////////////////////////
    // Allocate memory for output and init with 0
    ///////////////////////////////////////////////
    *out_re_pp = (double*)calloc(output_length, sizeof(double));
    *out_im_pp = (double*)calloc(output_length, sizeof(double));

    if (*out_re_pp == NULL || *out_im_pp == NULL)
        return -1;


    /////////////////////////////////////////
    // Do the filter operation on the noise
    /////////////////////////////////////////
    long int i;
    for (i = 0; i < len; i++) {
        long int j;

        // Even elements can be directly copied
        {
            long int pos = i * 2 + (4 * NUM_HB_COEFF);
            (*out_re_pp)[pos] += in_re_p[i];
            (*out_im_pp)[pos] += in_im_p[i];
        }

        // Add elements of the "positive" filter coefficients
        for (j = 0; j < NUM_HB_COEFF; j++) {
            long int pos = (i + j) * 2 + (1 + 4 * NUM_HB_COEFF);

            (*out_re_pp)[pos] += in_re_p[i] * halfBandFilterCoeff[j];
            (*out_im_pp)[pos] += in_im_p[i] * halfBandFilterCoeff[j];
        }

        // Add elements of the "negative" filter coefficients
        for (j = 0; j < NUM_HB_COEFF; j++) {
            long int pos = (i - j) * 2 + (-1 + 4 * NUM_HB_COEFF);

            (*out_re_pp)[pos] += in_re_p[i] * halfBandFilterCoeff[j];
            (*out_im_pp)[pos] += in_im_p[i] * halfBandFilterCoeff[j];
        }
    }

    return output_length;

}


long int filtered_noise(double **out_re_pp, double **out_im_pp,
        const long int len,
        const double bandwidth, const double variance) {

    // Output length of this filtered noise
    const long int output_len = len + (ROOT_RAISED_COSINE_NUM_TAPS - 1);

    ////////////////////////////////////////////////
    // Calculate filter raised cosine coefficients
    ////////////////////////////////////////////////
    double *filterCoeff = NULL;
    if (ROOT_RAISED_COSINE_NUM_TAPS != root_raised_cosine_filter(&filterCoeff, bandwidth,
            ROOT_RAISED_COSINE_ALPHA, ROOT_RAISED_COSINE_NUM_TAPS))
        return -1;

    ////////////////////////////////////////
    // Allocate memory and init with zeros
    ////////////////////////////////////////
    *out_re_pp = (double*)calloc(output_len, sizeof(double));
    *out_im_pp = (double*)calloc(output_len, sizeof(double));

    ////////////////////////////////////////////
    // Add filtered noise using a convolution
    // operation with the raised cosine filter
    // and set level
    ////////////////////////////////////////////
    long int i;
    const double sigma = sqrt(variance);
    for (i = 0; i < len; i++) {
        double noise_real, noise_imag;
        gaussian_noise(&noise_real, &noise_imag, 1);

        long int j;
        for (j = 0; j < ROOT_RAISED_COSINE_NUM_TAPS; j++) {
            (*out_re_pp)[i + j] += noise_real * filterCoeff[j] * sigma;
            (*out_im_pp)[i + j] += noise_imag * filterCoeff[j] * sigma;
        }
    }

    free (filterCoeff);

    return output_len;
}


long int generate_interferer(double **out_re_pp, double **out_im_pp,
        const unsigned int len,
        const double bandwidth, const double variance) {
    // This function is recursive, it calls itself with double
    // bandwidth and halved length and interpolates this afterwards.
    // The actual generation of the signal is done if the relative
    // bandwidth is >= 0.25

    if (bandwidth < 0.25) {

        // Here we call this function in a recursive way with double bandwidth
        // but halved length

        // Call this function
        double *inter_re_p = NULL;
        double *inter_im_p = NULL;
        long int inter_len = generate_interferer(&inter_re_p, &inter_im_p, len / 2,
                bandwidth * 2.0, variance);

        // Upsample signal by factor of 2
        long int out_len = upsample(inter_re_p, inter_im_p, out_re_pp, out_im_pp, inter_len);

        free (inter_re_p);
        free (inter_im_p);

        return out_len;
    }
    else {
        // Create the actual filtered noise
        return filtered_noise(out_re_pp, out_im_pp, len, bandwidth, variance);
    }
}
