/**
 *
 *    Copyright (C) 2017 J. Robert (FAU Erlangen-Nuernberg)
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 *    \brief This file contains the definitions for the filter functions.
 *
 */


#ifndef IEEE_LPWA_FILTER_H_
#define IEEE_LPWA_FILTER_H_


/**
 * Number of taps for the root-raised cosine filter. The
 * value must be odd. 151 results in an attenuation
 * of the unwanted components by 80 dB.
 */
#define ROOT_RAISED_COSINE_NUM_TAPS  151


/**
 * Factor for the root-raised cosine filter as defined
 * in the IEEE LPWA interference model
 */
#define ROOT_RAISED_COSINE_ALPHA     0.05



/**
 * \brief   Calculate root-raised cosine filter coefficients
 *
 * This function calculates the coefficients of a root-raised cosine filter. See
 * \sa{https://en.wikipedia.org/wiki/Root-raised-cosine_filter} for details
 * on the raised cosine filter.
 *
 * \param   filter_pp   Double pointer for allocating memory and storing the filter coefficients
 * \param   bandwidth   Relative (normalized to 1) bandwidth of the filter
 * \param   alpha       Raised cosine filter factor alpha
 * \param   num_taps    Number of taps of the filter
 *
 * \return  Number of taps, -1 on error
 */
long int root_raised_cosine_filter(double **filter_pp, const double bandwidth,
        const double alpha, const long int num_taps);

long int filtered_noise(double **out_re_pp, double **out_im_pp,
        const long int len,
        const double bandwidth, const double variance);

long int upsample(const double *in_re_p, const double *in_im_p,
        double **out_re_pp, double **out_im_pp,
        const long int len);

long int generate_interferer(double **out_re_pp, double **out_im_pp,
        const unsigned int len,
        const double bandwidth, const double variance);


#endif /* IEEE_LPWA_FILTER_H_ */
