/**
 *
 *    Copyright (C) 2017 J. Robert (FAU Erlangen-Nuernberg)
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.

 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 *    \brief This file contains the implementation for creating the playground.
 *
 */


#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

#include "ieee_lpwa_noise.h"
#include "ieee_lpwa_filter.h"
#include "ieee_lpwa_interferenceModel.h"



long int create_playground(double **out_re_pp, double **out_im_pp,
        const double duration_s,
        const double bandwidth_Hz, const double frequency_Hz,
        const interferer_layer_t* interferer_layers, const int num_interferer_layers,
        pathloss_model_callback pathloss_model, const double log_normal_fading_dB,
        const double min_tx_distance_m,
        const double noise_figure_dB) {

    // Return value for functions
    long int ret;

    // Number of samples required to store the requested bandwidth and length
    const long int num_output_samples = (long int)(duration_s * bandwidth_Hz);

    // Effective noise spectral density in dBm
    const double noise_spectral_density_dBmHz = THERMAL_NOISE_DENSITY_dBmHz + noise_figure_dB;

    // Effective noise spectral density in mW/Hz (linear scale)
    const double noise_spectral_density_lin = pow(10.0, noise_spectral_density_dBmHz / 10.0) * bandwidth_Hz;

    // Allocate memory and fill with thermal noise floor
    ret = noise_floor(out_re_pp, out_im_pp,  num_output_samples, noise_spectral_density_lin);
    if (ret < 0) {
        printf("Error while creating noise floor\n");
        return -1;
    }


    /////////////////////////
    // Loop over all layers
    /////////////////////////
    long int interferer_layer;
    for (interferer_layer = 0; interferer_layer < num_interferer_layers; interferer_layer++) {


        ////////////////////////////////////////////
        // Check radius that we have to consider
        // Only signals above the thermal noise
        // floor have to be considered. Furthermore
        // we give an additional margin of 3dB.
        ////////////////////////////////////////////

        // Spectral density of the interferer in dBm/Hz
        const double signal_spectral_density_dBmHz =
                interferer_layers[interferer_layer].tx_power_dBm
                -10.0 * log10(interferer_layers[interferer_layer].bandwidth_Hz);


        // Additional margin for potential log-normal fading
        const double log_normal_margin_dB =
                log_normal_fading_dB * 4.0; // 4 times the standard deviation covers 0.999936657516334 of all cases

        // Check if we have to consider this layer or if it fully disappears in the noise
        if (signal_spectral_density_dBmHz + log_normal_margin_dB - pathloss_model(min_tx_distance_m, frequency_Hz)
                            < noise_spectral_density_dBmHz - 3.0)
            continue; // Contninue to next layer if completely within thermal noise


        // Estimated radius in m we have to consider
        double maxRadius_m = min_tx_distance_m;

        // If radius zero assume at least 10m to be able to multiply the radius in the while loop
        if (maxRadius_m == 0)
            maxRadius_m = 10;

        // Estimate radius iteratively, multiply by 1.1 if radius still too small
        while (signal_spectral_density_dBmHz + log_normal_margin_dB - pathloss_model(maxRadius_m, frequency_Hz)
                > noise_spectral_density_dBmHz - 3.0) {
            maxRadius_m *= 1.1;
        }

#ifndef NDEBUG
        printf("Layer %i: Max radius %.2fm\n", (int)interferer_layer, (float)maxRadius_m);
#endif

        // Calculate size of the covered area in km^2
        const double area_km =
                M_PI * (maxRadius_m / 1000.0) * (maxRadius_m / 1000.0)
                - M_PI * (min_tx_distance_m / 1000.0) * (min_tx_distance_m / 1000.0);

        // Calculate active number of interferer in 1/s
        const double lambda =
                interferer_layers[interferer_layer].arrival_rate / 1.0e6
                * bandwidth_Hz / 1.0e6
                * area_km;

        /////////////////////////////////////
        // Add interferer
        /////////////////////////////////////

        // Counter for the number of interferer
        long int num_interferer = 0;

        // Current sample position within playground
        // We start at negative time to avoid ramp-up problems.
        // Length * bandwidth * 2 ensures start that is early enough to
        // avoid problems caused by the start phase of the filters
        long int current_sample_pos = -interferer_layers[interferer_layer].length_s * bandwidth_Hz * 2.0;


        // Loop while start of interferer still within the playground
        while (current_sample_pos < num_output_samples) {

            // Calculate offset to start of next interferer. As the interferer are Poisson distributed,
            // the inter-arrival times follow an exponential distribution.
            // See \sa{https://en.wikipedia.org/wiki/Exponential_distribution} for details.
            long int offset = -1.0 / lambda * log(rand_uniform());
            current_sample_pos += offset;


            // Leave the loop if we are outside the playground, we are done
            if (current_sample_pos >= (int)num_output_samples)
                break;

            ///////////////////////////////
            // Calculate distance TX->RX
            ///////////////////////////////
            double distance;            // Distance TX->RX in m

            // Loop until we have a value inside the valid range
            do {
                double x = rand_uniform() * maxRadius_m;
                double y = rand_uniform() * maxRadius_m;

                distance = sqrt(x*x + y*y);

            } while (distance < min_tx_distance_m && min_tx_distance_m > maxRadius_m);


            // Actual channel fading in dB
            double channel_dB = pow(10.0, -pathloss_model(distance, frequency_Hz) / 20.0);

            //////////////////////////////////////////////
            // Calculate log-normal fading (if required)
            //////////////////////////////////////////////
            double log_normal_dB = 0;
            if (log_normal_fading_dB > 0) {
                gaussian_noise_real(&log_normal_dB, 1);
                log_normal_dB *= log_normal_fading_dB;
            }

            // Check if the signal is higher than the threshold, if not we do not have to calculate
            if (signal_spectral_density_dBmHz + log_normal_dB - pathloss_model(distance, frequency_Hz)
                          < noise_spectral_density_dBmHz - 3.0)
                continue;


            num_interferer++;

            ///////////////////////////////
            // Actually create interferer
            ///////////////////////////////

            double *inter_re_p = NULL;  // Pointer storing real part of interferer signal
            double *inter_im_p = NULL;  // Pointer storing imaginary part of interferer signal

            // Allocate memory and create interferer
            long int len = generate_interferer(&inter_re_p, &inter_im_p,
                    interferer_layers[interferer_layer].length_s * bandwidth_Hz,
                    interferer_layers[interferer_layer].bandwidth_Hz / bandwidth_Hz,
                    1.0);


            long int i;
            for (i = 0; i < len; i++) {
                inter_re_p[i] *= channel_dB;
                inter_im_p[i] *= channel_dB;
            }

            // Calculate random frequency offset
            double freqOffset = rand_uniform();
            double phase = 0;


            for (i = 0; i < len; i++) {
                // Check if in valid range
                if (i + current_sample_pos < 0)
                    continue;
                if (i + current_sample_pos >= num_output_samples)
                    break;

                double real_phase = cos(2 * M_PI * phase);
                double imag_phase = sin(2 * M_PI * phase);

                (*out_re_pp)[i + current_sample_pos] += inter_re_p[i] * real_phase;
                (*out_re_pp)[i + current_sample_pos] -= inter_im_p[i] * imag_phase;
                (*out_im_pp)[i + current_sample_pos] += inter_im_p[i] * real_phase;
                (*out_im_pp)[i + current_sample_pos] += inter_re_p[i] * imag_phase;

                phase += freqOffset;
            }

            free (inter_re_p);
            free (inter_im_p);


        }
#ifndef NDEBUG
        printf("Number of interferer: %i\n", (int)num_interferer);
#endif
    }

    return num_output_samples;
}

void free_playground(double **out_re_pp, double **out_im_pp) {
    free (*out_re_pp);
    free (*out_im_pp);
}

