/**
 *
 *    Copyright (C) 2017 J. Robert (FAU Erlangen-Nuernberg)
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.

 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 *    \brief This file contains the definitions for the playground functions.
 *
 */


#ifndef IEEE_LPWA_INTERFERENCEMODEL_H_
#define IEEE_LPWA_INTERFERENCEMODEL_H_

#include "ieee_lpwa_pathlossModel.h"

//! Thermal noise density in dBm/Hz
#define THERMAL_NOISE_DENSITY_dBmHz     -174.0


/**
 * \brief   Struct that defines the different interferer layers
 */
typedef struct interferer_layer {
    double arrival_rate;    //!< Arrival rate in arrivals per seconds per MHz per km^2
    double length_s;        //!< Length of interferer in s
    double bandwidth_Hz;    //!< Bandwidth of interferer in Hz
    double tx_power_dBm;    //!< Transmit power in dBm
} interferer_layer_t;


/**
 * \brief Create playground
 *
 * This function creates the complete playground, i.e. the thermal noise floor and the interferers.
 * The definition of the playground is given in the corresponding parameters.
 * The playground allocates the memory itself, which has to be freed after use.
 *
 *
 * \param out_re_pp         Double pointer to return allocated memory and real part of playground
 * \param out_im_pp         Double pointer to return allocated memory and imaginary part of playground
 *
 * \param duration_s        Duration of the playground in s
 * \param bandwidth_Hz      Bandwidth of playground in Hz
 * \param frequency_Hz      Centre frequency of playground in Hz
 *
 * \param interferer_layers         Pointer to structs describing the interferer layers
 * \param num_interferer_layers     Number of interferer layers
 *
 * \param pathloss_model            Callback function for pathloss model
 * \param log_normal_fading_dB      Log normal fading of the pathloss model
 *
 * \param min_tx_distance_m Minimum distance between transmitters and the receiver station
 * \param noise_figure_dB   Noise figure
 *
 * \return Size of the playground in complex samples
 */
long int create_playground(double **out_re_pp, double **out_im_pp,
        const double duration_s,
        const double bandwidth_Hz, const double frequency_Hz,
        const interferer_layer_t* interferer_layers, const int num_interferer_layers,
        pathloss_model_callback pathloss_model, const double log_normal_fading_dB,
        const double min_tx_distance_m,
        const double noise_figure_dB);


/**
 * \brief Free playground memory
 */
void free_playground(double **out_re_pp, double **out_im_pp);

#endif /* IEEE_LPWA_INTERFERENCEMODEL_H_ */

