/**
 *
 *    Copyright (C) 2017 J. Robert (FAU Erlangen-Nuernberg)
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 *    \brief This file contains the required definitions for the noise functions.
 *
 */


#ifndef IEEE_LPWA_NOISE_H_
#define IEEE_LPWA_NOISE_H_

#include "rand_generator.h"

/**
 * \brief   Calculate uniform random variable in [0,1)
 *   
 * This is just a wrapper for the genrand_real2() function defined in mt19937ar.c
 *
 * \return  Random value in [0,1)
 */
static inline double rand_uniform(void) {
    return genrand_real2();
}


/**
 * \brief   Calculate complex Gaussian noise
 *
 * This function creates random values derived from a complex Gaussian
 * distribution with zero mean and variance one using the Box-Muller method,
 * see \sa{https://en.wikipedia.org/wiki/Box%E2%80%93Muller_transform}
 * for details.
 * The arrays must be able to store the requested num variables.
 *
 * \param   data_re_p   Pointer to existing array for real part
 * \param   data_im_p   Pointer to existing array for imaginary part
 * \param   num         Number of requested random variables, -1 on error
 *
 */
void gaussian_noise(double *data_re_p, double *data_im_p, const long int num);

/**
 * \brief   Calculate real Gaussian noise
 *
 * This function creates random values derived from a Gaussian
 * distribution with zero mean and variance one using the Box-Muller method,
 * see \sa{https://en.wikipedia.org/wiki/Box%E2%80%93Muller_transform}
 * for details.
 * The arrays must be able to store the requested num variables.
 *
 * \param   data_p      Pointer to existing array for real part
  * \param   num         Number of requested random variables, -1 on error
 *
 */
void gaussian_noise_real(double *data_p, const long int num);


/**
 * \brief   Create noise floor with Gaussian noise
 *
 * This function allocates memory for data_re and data_im and fills
 * the memory with complex Gaussian noise with requested variance
 * and zero mean.
 *
 * The calling function has to free the memory.
 *
 * \param   data_re_pp      Double pointer for allocating memory and storing the real part
 * \param   data_im_pp      Double pointer for allocating memory and storing the imaginary part
 * \param   num             Number of requested random variables
 * \param   variance_lin    Variance of the Gaussian noise in linear scale
 *
 * \return  Number of return random variables
 *
 */
long int noise_floor(double **data_re_pp, double **data_im_pp,
        const unsigned int num, double variance_lin);




#endif /* IEEE_LPWA_NOISE_H_ */
