/**
 *
 *    Copyright (C) 2017 J. Robert (FAU Erlangen-Nuernberg)
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.

 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 *    \brief This file contains the implementation of the path-loss models.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

#include "ieee_lpwa_pathlossModel.h"

double outdoor_rural_32m(const double distance, const double frequency) {
    const double h_bs = 32.0;       //!< Base station height in m
    const double h_ms = 1.5;        //!< Mobile station height in m

    const double PL = (44.9 - 6.55 * log10(h_bs))*log10(distance / 1000.0) + 45.5
            +(35.46 - 1.1 *h_ms) * log10(frequency / 1.0e6) - 13.82 * log10(h_bs) + 0.7 * h_ms;

    return PL;
}


double outdoor_rural_140m(const double distance, const double frequency) {
    const double h_bs = 140.0;      //!< Base station height in m
    const double h_ms = 1.5;        //!< Mobile station height in m

    const double PL = (44.9 - 6.55 * log10(h_bs))*log10(distance / 1000.0) + 45.5
            +(35.46 - 1.1 *h_ms) * log10(frequency / 1.0e6) - 13.82 * log10(h_bs) + 0.7 * h_ms;

    return PL;
}


double outdoor_urban_32m(const double distance, const double frequency) {
    const double h_bs = 32.0;       //!< Base station height in m
    const double h_ms = 1.5;        //!< Mobile station height in m

    const double PL = (44.9 - 6.55 * log10(h_bs))*log10(distance / 1000.0) + 45.5
            +(35.46 - 1.1 *h_ms) * log10(frequency / 1.0e6) - 13.82 * log10(h_bs) + 0.7 * h_ms
            + 3.0;

    return PL;
}


double outdoor_urban_140m(const double distance, const double frequency) {
    const double h_bs = 140.0;      //!< Base station height in m
    const double h_ms = 1.5;        //!< Mobile station height in m

    const double PL = (44.9 - 6.55 * log10(h_bs))*log10(distance / 1000.0) + 45.5
            +(35.46 - 1.1 *h_ms) * log10(frequency / 1.0e6) - 13.82 * log10(h_bs) + 0.7 * h_ms
            + 3.0;

    return PL;
}


double outdoor_device_to_device(const double distance, const double frequency) {
    if (frequency != 900.0e6) {
        printf("outdoor_device_to_device only specified for 900 MHz!\n");
    }
    const double PL = -6.17 + 58.6 * log10(distance);
    return PL;
}


double indoor(const double distance, const double frequency) {

    const double d_BP = 5.0;        //!< Breakpoint distance in m
    double PL;

    if (distance <= d_BP) {
        // Pathloss before breakpoint (line of sight)
        PL = 20.0 * log10((4.0 * M_PI / 300000000.0) * frequency * distance);
    }
    else {
        // Pathloss after breakpoint
        PL = 20.0 * log10((4.0 * M_PI / 300000000.0) * frequency * d_BP);
        PL += 35.0 * log10(distance / d_BP);
    }
    return PL;
}

