/**
 *
 *    Copyright (C) 2017 J. Robert (FAU Erlangen-Nuernberg)
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.

 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 *    \brief This file contains the definitions for the path-loss models.
 *
 */

#ifndef IEEE_LPWA_PATHLOSSMODEL_H_
#define IEEE_LPWA_PATHLOSSMODEL_H_

/**
 * Definition of callback function to change propagation channel model
 */
typedef double(*pathloss_model_callback)(const double, const double);



/**
 * Path loss model for outdoor rural with base-station antenna height of 32m
 *
 * The model uses the propagation model as defined in 3GPP TR 25.996 Table 5.1
 *
 * \param   distance    Distance between TX and RX in [m]
 * \param   frequency   Signal frequency in [Hz]
 */
double outdoor_rural_32m(const double distance, const double frequency);


/**
 * Path loss model for outdoor rural with base-station antenna height of 140m
 *
 * The model uses the propagation model as defined in 3GPP TR 25.996 Table 5.1
 *
 * \param   distance    Distance between TX and RX in [m]
 * \param   frequency   Signal frequency in [Hz]
 */
double outdoor_rural_140m(const double distance, const double frequency);


/**
 * Path loss model for outdoor urban with base-station antenna height of 32m
 *
 * The model uses the propagation model as defined in 3GPP TR 25.996 Table 5.1
 *
 * \param   distance    Distance between TX and RX in [m]
 * \param   frequency   Signal frequency in [Hz]
 */
double outdoor_urban_32m(const double distance, const double frequency);


/**
 * Path loss model for outdoor urban with base-station antenna height of 140m
 *
 * The model uses the propagation model as defined in 3GPP TR 25.996 Table 5.1
 *
 * \param   distance    Distance between TX and RX in [m]
 * \param   frequency   Signal frequency in [Hz]
 */
double outdoor_urban_140m(const double distance, const double frequency);



/**
 * Path loss model for outdoor to outdoor
 *
 * The model usses the IEEE 802.11 TGah Channel Model, IEEE 802.11-11/968r4, Sec. 3.3
 *
 * \param   distance    Distance between TX and RX in [m]
 * \param   frequency   Signal frequency in [Hz]
 */
double outdoor_device_to_device(const double distance, const double frequency);


/**
 * Path loss model for indoor propagation
 *
 * The model usses the IEEE 802.11 TGah Channel Model, IEEE 802.11-11/968r4, Sec. 3.5, Model A
 *
 *
 * \param   distance    Distance between TX and RX in [m]
 * \param   frequency   Signal frequency in [Hz]
 */
double indoor(const double distance, const double frequency);





#endif /* IEEE_LPWA_PATHLOSSMODEL_H_ */
