/**
 *
 *    Copyright (C) 2017 J. Robert (FAU Erlangen-Nuernberg)
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.

 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 *    \brief This file implements a simple example to demonstrate the use the channel model.
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>

#include "ieee_lpwa_interferenceModel.h"
#include "ieee_lpwa_noise.h"
#include "ieee_lpwa_filter.h"
#include "ieee_lpwa_pathlossModel.h"
#include "rand_generator.h"

/**
 * \brief   Show help for this program
 */
void show_help() {
    printf("Help\n");
    printf("-h      Shows this help\n");
    printf("-p arg  Path loss model: outdoor_rural_32m, outdoor_rural_140m,\n"
           "        outdoor_urban_32m, outdoor_urban_140m, outdoor_device_to_device,\n"
           "        indoor (default)\n");
    printf("-b arg  Bandwidth in Hz, default 1 MHz\n");
    printf("-d arg  Duration in s, default 1 s\n");
    printf("-f arg  Centre frequency in Hz, default 900 MHz\n");
    printf("-l arg  Log-normal fading in dB, default 0 dB\n");
    printf("-n arg  Noise figure in dB, default 0 dB\n");
    printf("-m arg  Min distance to TX in m, default 0 m\n");
    printf("-a arg  Acticity multiplier, default is 1.0\n");
    printf("-o arg  Output file name, default is out.bin\n"
           "        The ouput file format is interleaved complex double\n");
    printf("-r arg  Random seed, default is init with current time\n");
    printf("-s arg  Create set, i.e. multiple playgrounds, default is off\n");
    printf("-e      Use float precision instead of double for output encoding\n");


    return;
}


/**
 * \brief Main function for simple test program
 */
int main(int argc, char **argv) {
    printf("=======================================================================\n");
    printf("IEEE 802.15 IG LPWA Interference Model, Copyright (C) 2017 Joerg Robert\n");
    printf("This program comes with ABSOLUTELY NO WARRANTY.\n");
    printf("This is free software, and you are welcome to redistribute it\n");
    printf("under certain conditions; see file gpl.txt for details.\n");
    printf("\n");
    printf("Use -h for help\n");
    printf("=======================================================================\n");

    /**
     * Program options
     */
    pathloss_model_callback pathloss_model = indoor;
    const char *pathloss_model_name = "indoor";
    double bandwidth_Hz = 1.0e6;
    double frequency_Hz = 900.0e6;
    double log_normal_fading_dB = 0;
    double noise_figure_dB = 0;
    double duration_s = 1.0;
    double min_tx_distance_m = 0;
    double activity_factor = 1.0;
    const char *output_file_name = "out.bin";
    unsigned long random_seed = time(NULL);
    unsigned int set_size = 0;
    int f32 = 0;


    int error = 0;
    int arg;

    for (arg = 1 ;arg < argc; arg++) {
        // Check for "-"
        if (argv[arg][0] != '-') {
            error = 1;
            break;
        }

        // Get command
        int command = argv[arg][1];

        // Set optional arg if present
        const char *optarg = "";

        if (arg + 1 < argc)
            optarg = argv[arg + 1];


        switch (command) {

        case 'h': {     // Show help
            show_help();
            return 0;
        } break;

        case 'p': {     // Set pathloss model
            arg++;
            if (strcmp(optarg, "outdoor_rural_32m") == 0) {
                pathloss_model = outdoor_rural_32m;
                pathloss_model_name = "outdoor_rural_32m";
                break;
            }
            if (strcmp(optarg, "outdoor_rural_140m") == 0) {
                pathloss_model = outdoor_rural_140m;
                pathloss_model_name = "outdoor_rural_140m";
                break;
            }
            if (strcmp(optarg, "outdoor_urban_32m") == 0) {
                pathloss_model = outdoor_urban_32m;
                pathloss_model_name = "outdoor_urban_32m";
                break;
            }
            if (strcmp(optarg, "outdoor_urban_140m") == 0) {
                pathloss_model = outdoor_urban_140m;
                pathloss_model_name = "outdoor_urban_140m";
                break;
            }
            if (strcmp(optarg, "outdoor_device_to_device") == 0) {
                pathloss_model = outdoor_device_to_device;
                pathloss_model_name = "outdoor_device_to_device";
                break;
            }
            if (strcmp(optarg, "indoor") == 0) {
                pathloss_model = indoor;
                pathloss_model_name = "indoor";
                break;
            }
            printf("Unkown pathloss model. Use -h for help.\n");
            return 0;

        } break;

        case 'b': {     // Bandwidth
            arg++;
            bandwidth_Hz = strtod(optarg, NULL);
        } break;

        case 'f': {     // Frequency
            arg++;
            frequency_Hz = strtod(optarg, NULL);
        } break;

        case 'l': {     // Lognormal fading
            arg++;
            log_normal_fading_dB = strtod(optarg, NULL);
        } break;

        case 'n': {     // Noise figure
            arg++;
            noise_figure_dB = strtod(optarg, NULL);
        } break;

        case 'd': {     // Duation
            arg++;
            duration_s = strtod(optarg, NULL);
        } break;

        case 'm': {     // Minimum distance between TX and RX
            arg++;
            min_tx_distance_m = strtod(optarg, NULL);
        } break;

        case 'a': {     // Activity factor
            arg++;
            activity_factor = strtod(optarg, NULL);
        } break;

        case 'o': {     // Output filename
            arg++;
            output_file_name = optarg;
        } break;

        case 'r': {     // Random seed
            arg++;
            random_seed = strtod(optarg, NULL);
        } break;

        case 's': {     // Create set
            arg++;
            set_size = strtod(optarg, NULL);
        } break;

        case 'e': {     // Use float as output encoding
            f32 = 1;
        } break;


        default: {     // Unknown option
            error = 1;
        } break;
        }
    }

    if (error) {
        printf("Unknown option or missing parameter, use -h for help\n");
        return -1;
    }



    // Set random seed
    init_genrand (random_seed);


    // Definition according to https://mentor.ieee.org/802.15/documents document IEEE 802.15-17-0037-01-lpwa, slide 12
    const interferer_layer_t interferer[4] = {
            {activity_factor * 0.8, 0.005, 100.0e3, 10},
            {activity_factor * 0.15, 0.005, 20.0e3, 10},
            {activity_factor * 0.04, 0.100, 10.0e3, 10},
            {activity_factor * 0.01, 0.030, 1000.0e3, 10},
    };


    ////////////////////////////////////
    // Write config parameters to file
    ////////////////////////////////////
    {
 
        // Create tmp output file name
        char tmp_config_file_name[1024];
        sprintf(tmp_config_file_name, "%s.txt", output_file_name);

        FILE *configOutFile;
        configOutFile = fopen(tmp_config_file_name, "w"); 

        fprintf(configOutFile, "-p %s ", pathloss_model_name);
        fprintf(configOutFile, "-b %f ", bandwidth_Hz);
        fprintf(configOutFile, "-d %f ", duration_s);
        fprintf(configOutFile, "-f %f ", frequency_Hz);
        fprintf(configOutFile, "-l %f ", log_normal_fading_dB);
        fprintf(configOutFile, "-n %f ", noise_figure_dB);
        fprintf(configOutFile, "-m %f ", min_tx_distance_m);
        fprintf(configOutFile, "-a %f ", activity_factor);
        fprintf(configOutFile, "-o %s ", output_file_name);
        fprintf(configOutFile, "-r %lu ", random_seed);
        fprintf(configOutFile, "-s %i ", set_size);
        if (f32)
            fprintf(configOutFile, "-e ");
        fclose(configOutFile);
    }
  
    ////////////////////////////////
    // Itereate over set
    ////////////////////////////////
    unsigned int set_idx;
    for (set_idx = 0; set_idx < set_size || set_idx == 0; set_idx++) {
        // Memory for storing data
        double *real_data, *imag_data;

        printf("Creating playground ...");
        fflush(stdout);

        int len = create_playground(&real_data, &imag_data,
                duration_s,
                bandwidth_Hz, frequency_Hz,
                interferer, sizeof(interferer) / sizeof(interferer_layer_t),
                pathloss_model, log_normal_fading_dB,
                min_tx_distance_m,
                noise_figure_dB);


        ///////////////////////
        // Write data to file
        ///////////////////////
        FILE* outFile;

        // Create tmp output file name
        char tmp_output_file_name[1024];
        if (set_size > 0) {
            sprintf(tmp_output_file_name, "%s_%i", output_file_name, set_idx);
        }
        else {
            strcpy(tmp_output_file_name, output_file_name);
        }

        outFile = fopen(tmp_output_file_name, "wb");

        
        if (f32) { // store using float precision
            int i;
            for (i = 0; i < len; i++) {
                float real_float = real_data[i];
                float imag_float = imag_data[i];
                fwrite((void*)&real_float, sizeof(float), 1, outFile);
                fwrite((void*)&imag_float, sizeof(float), 1, outFile);
            }
        }
        else { // store using double precision
            int i;
            for (i = 0; i < len; i++) {
                fwrite((void*)&real_data[i], sizeof(double), 1, outFile);
                fwrite((void*)&imag_data[i], sizeof(double), 1, outFile);
            }
        }

        fclose(outFile);

        // Free memory
        free_playground(&real_data, &imag_data);

        printf("done! Written to file: %s\n", tmp_output_file_name);
        fflush(stdout);
    }

    return 0;
}

